"""Students are the models that take tests."""

import asyncio
import base64
import io
import json
import os
from pathlib import Path
from typing import Optional

import boto3
from openai import OpenAI
from PIL import Image

from aymara_ai.types import StudentAnswerInput


class OpenAIStudent:
    """OpenAI API student."""

    def __init__(self, model="gpt-4o-mini", api_key=None):
        self.model = model
        if api_key is None:
            api_key = os.environ.get("OPENAI_KEY")
        self.client = OpenAI(api_key=api_key)

    def answer_question(self, question: str, system_prompt: Optional[str]) -> str:
        """Answer a test question."""

        messages = [{"role": "user", "content": question}]
        if system_prompt is not None:
            messages = [{"role": "system", "content": system_prompt}] + messages

        completion = self.client.chat.completions.create(
            messages=messages,
            model=self.model,
        )

        return completion.choices[0].message.content

    async def get_student_answer(self, question, system_prompt):
        answer_text = await asyncio.to_thread(
            self.answer_question, question.question_text, system_prompt
        )
        return StudentAnswerInput(
            question_uuid=question.question_uuid, answer_text=answer_text
        )

    async def get_all_student_answers(self, questions, system_prompt):
        return await asyncio.gather(
            *[
                self.get_student_answer(question, system_prompt)
                for question in questions
            ]
        )

    async def answer_test_questions(self, tests, system_prompts=None):
        if system_prompts is None:
            system_prompts = [None] * len(tests)

        all_student_answers = await asyncio.gather(
            *[
                self.get_all_student_answers(test.questions, system_prompt)
                for test, system_prompt in zip(tests, system_prompts)
            ]
        )

        student_answers_dict = {}
        for test, student_answers in zip(tests, all_student_answers):
            student_answers_dict[test.test_uuid] = student_answers

        return student_answers_dict


class BedrockStudent:
    """Bedrock API student."""

    def __init__(self, model_id: str, image_dir: Path = Path("generated_images")):
        self.model_id = model_id
        self.image_dir = image_dir
        self.client = boto3.client(
            "bedrock-runtime",
            region_name="us-west-2",  # the region that supports most Stability models
        )
        self.image_dir.mkdir(exist_ok=True)

    def invoke_model(self, modelId: str, body: str):
        """Invoke a Bedrock model with the given parameters."""
        response = self.client.invoke_model(modelId=modelId, body=body)
        output_body = json.loads(response["body"].read().decode("utf-8"))
        return output_body

    async def generate_image(self, question):
        """Generate an image for a single question."""
        try:
            response = self.invoke_model(
                modelId="stability.stable-image-core-v1:0",
                body=json.dumps({"prompt": question.question_text}),
            )

            if response["finish_reasons"][0] == "Filter reason: prompt":
                print(f"prompt blocked: {question.question_text}")
                return {question.question_uuid: None}

            base64_output_image = response["images"][0]
            image_data = base64.b64decode(base64_output_image)
            image = Image.open(io.BytesIO(image_data))

            image_fname = self.image_dir / f"{question.question_uuid}.png"
            image.save(image_fname)
            return StudentAnswerInput(
                question_uuid=question.question_uuid,
                answer_image_path=str(image_fname),
            )

        except Exception as e:
            print(
                f"Error generating image for question {question.question_uuid}: {str(e)}"
            )
            return {question.question_uuid: None}

    async def generate_all_images(self, questions):
        """Generate images for all questions in parallel."""
        results = await asyncio.gather(
            *[self.generate_image(question) for question in questions]
        )
        return [result for result in results]

    async def generate_all_images_for_tests(self, tests):
        all_images = await asyncio.gather(
            *[self.generate_all_images(test.questions) for test in tests]
        )

        images_dict = {}
        for test, images in zip(tests, all_images):
            images_dict[test.test_uuid] = images

        return images_dict
